#include "defs.h"
#include "llist.h"

#include <string.h>
#include <stdio.h>


// Testing
#define PRINT_DEBUG 1


// Local macros
#if PRINT_DEBUG
#  define PRINT_DEBUG_PRINT(x, y...) { fprintf(stderr, x, ##y); fflush(stderr); }
#else
#  define PRINT_DEBUG_PRINT(x, y...) {}
#endif
#define PRINT_ERROR_PRINT(x, y...) { fprintf(stderr, "LIST_ERROR: " x, ##y); fflush(stderr); }


// Prototypes
PCHAR return_start_flag(UINT32 FlagKey);
PCHAR return_end_flag(UINT32 FlagKey);


// Globals
extern FILE *fpfinal;
extern UINT32 fpfinalcount;
STRING_TABLE start_flags[] = {
	{0x01     ,  "<ERR,B,0x01>"  },
	{0x02     ,  "<ERR,B,0x02>"  },
	{0x04     ,  "<ERR,B,0x04>"  },
	{0x08     ,  "<ERR,B,0x08>"  },
	{0x10     ,  "<ERR,B,0x10>"  },
	{0x20     ,  "<ERR,B,0x20>"  },

	{Cw_Flag  ,  "<w>"  },
	{Ck_Flag  ,  "<k>"  },
	{Cy_Flag  ,  "<y>"  },
	{Cm_Flag  ,  "<m>"  },
	{Cc_Flag  ,  "<c>"  },
	{Cb_Flag  ,  "<b>"  },
	{Cg_Flag  ,  "<g>"  },
	{Cr_Flag  ,  "<r>"  },

	{S9_Flag  ,  "<9>"  },
	{S8_Flag  ,  "<8>"  },
	{S7_Flag  ,  "<7>"  },
	{S6_Flag  ,  "<6>"  },
	{S5_Flag  ,  "<5>"  },
	{S4_Flag  ,  "<4>"  },
	{S3_Flag  ,  "<3>"  },
	{S2_Flag  ,  "<2>"  },
	{S1_Flag  ,  "<1>"  },
	{S0_Flag  ,  "<0>"  },

	{U3_Flag  ,  "<U>"  },
	{U2_Flag  ,  "<U>"  },
	{U1_Flag  ,  "<U>"  },

	{TT_Flag  ,  "<TT>"  },
	{S_Flag   ,  "<S>"  },
	{I_Flag   ,  "<I>"  },
	{EM_Flag  ,  "<EM>" },
	{B_Flag   ,  "<B>"  }
};
UINT32 start_flags_size = sizeof(start_flags) / sizeof(start_flags[0]);

STRING_TABLE end_flags[] = {
	{0x01     ,  "<ERR,E,0x01>"  },
	{0x02     ,  "<ERR,E,0x02>"  },
	{0x04     ,  "<ERR,E,0x04>"  },
	{0x08     ,  "<ERR,E,0x08>"  },
	{0x10     ,  "<ERR,E,0x10>"  },
	{0x20     ,  "<ERR,E,0x20>"  },

	{Cw_Flag  ,  "</w>"  },
	{Ck_Flag  ,  "</k>"  },
	{Cy_Flag  ,  "</y>"  },
	{Cm_Flag  ,  "</m>"  },
	{Cc_Flag  ,  "</c>"  },
	{Cb_Flag  ,  "</b>"  },
	{Cg_Flag  ,  "</g>"  },
	{Cr_Flag  ,  "</r>"  },

	{S9_Flag  ,  "</9>"  },
	{S8_Flag  ,  "</8>"  },
	{S7_Flag  ,  "</7>"  },
	{S6_Flag  ,  "</6>"  },
	{S5_Flag  ,  "</5>"  },
	{S4_Flag  ,  "</4>"  },
	{S3_Flag  ,  "</3>"  },
	{S2_Flag  ,  "</2>"  },
	{S1_Flag  ,  "</1>"  },
	{S0_Flag  ,  "</0>"  },

	{U3_Flag  ,  "</U>"  },
	{U2_Flag  ,  "</U>"  },
	{U1_Flag  ,  "</U>"  },

	{TT_Flag  ,  "</TT>"  },
	{S_Flag   ,  "</S>"  },
	{I_Flag   ,  "</I>"  },
	{EM_Flag  ,  "</EM>" },
	{B_Flag   ,  "</B>"  }
};
UINT32 end_flags_size = sizeof(end_flags) / sizeof(end_flags[0]);


// Utilities
PCHAR
return_start_flag(UINT32 FlagKey)
{
	int index;
	
	for (index=31; index >= 0; index--) {
		if (FlagKey & (1<<index)) {
			//PRINT_DEBUG_PRINT("Returning %s\n", start_flags[index].String);
			return(start_flags[index].String);
		}
	}

	return("<ERROR>");
}

PCHAR
return_end_flag(UINT32 FlagKey)
{
	UINT32 index;
	
	for (index=31; index >= 0; index--) {
		if (FlagKey & (1<<index)) {
			//PRINT_DEBUG_PRINT("Returning %s\n", end_flags[index].String);
			return(end_flags[index].String);
		}
	}

	return("</ERROR>");
}

UINT32
count_changes(UINT32 FlagKey)
{
	int index, count = 0;
	
	for (index=31; index >= 0; index--) {
		if (FlagKey & (1<<index)) {
			count++;
		}
	}

	return count;
}

//
//
// Version 2.0
//
//


UINT32
find_changes(PLIST_ENTRY ListEntry)
{
	UINT32 diffState = 0;
	UINT32 counter = 0;
	UINT32 done = 0;

	diffState = (get_previous_state(ListEntry) ^ get_this_state(ListEntry));

	return diffState;
}

UINT32
should_add_color(UINT32 turnons, PLIST outputStack)
{
	PLIST_ENTRY	tempPtr2 = get_last_entry(outputStack);
	UINT32 k;
		
	if (!(turnons & COLOR_FLAG_MASK))
		return turnons;

	while (tempPtr2) {
		k = (get_this_state(tempPtr2) & COLOR_FLAG_MASK);
		if (k) {
			if (k & turnons) {
				turnons &= ~k;	
			}
			break;
		}
		tempPtr2 = get_previous_entry(tempPtr2);
	}
	return turnons;
}

UINT32
should_add_size(UINT32 turnons, PLIST outputStack)
{
	PLIST_ENTRY	tempPtr2 = get_last_entry(outputStack);
	UINT32 k;
		
	if (!(turnons & SIZE_FLAG_MASK))
		return turnons;

	while (tempPtr2) {
		k = (get_this_state(tempPtr2) & SIZE_FLAG_MASK);
		if (k) {
			if (k & turnons) {
				turnons &= ~k;	
			}
			break;
		}
		tempPtr2 = get_previous_entry(tempPtr2);
	}
	return turnons;
}

void
look_ahead(PLIST_ENTRY ListEntry, UINT32 turnons, PLIST outputStack)
{
	PLIST_ENTRY	tempPtr = ListEntry;
	PLIST newstack;
	int i;
	UINT32 k, state;

	new_list(&newstack);

	while (turnons && tempPtr) {
		k = (~get_this_state(tempPtr) & turnons);
		if (k) {
			for (i = 0; i < 32; i++) {
				if (k & (1<<i)) {
					new_record_special(newstack, (1<<i));
					turnons &= ~(1<<i);
				}
			}
		}
		tempPtr = get_next_entry(tempPtr);
	}

	if (turnons) {
		for (i = 0; i < 32; i++) {
			if (turnons & (1<<i)) {
				new_record_special(newstack, (1<<i));
			}
		}
	}

	tempPtr = get_last_entry(newstack);
	while (tempPtr) {
		state = get_this_state(tempPtr);
		fpfinalcount += strlen(return_start_flag(state));
		fputs(return_start_flag(state), fpfinal);
		new_record_special(outputStack, state);
		tempPtr = get_previous_entry(tempPtr);
	}
	release_list(newstack);
}

void
process_entry(PLIST_ENTRY ListEntry, PLIST OutputStack)
{
	UINT32 mask, turnons, turnoffs;
	UINT32 index;

//fprintf(stderr, "Processing: 0x%08p\n", ListEntry);
//print_entry(ListEntry);
//fprintf(stderr, "OutputStack: 0x%08p\n", ListEntry);
//print_list(OutputStack);

	// figure out what has changed in this transition
	mask = find_changes(ListEntry);

	turnoffs = get_previous_state(ListEntry) & mask;
	turnons = ~get_previous_state(ListEntry) & mask;

	// For Color and Sizes
	// Don't turn of color or size unless we are going to zero state

	if (get_this_state(ListEntry) & SIZE_FLAG_MASK)
		turnoffs &= ~SIZE_FLAG_MASK;

	if (get_this_state(ListEntry) & COLOR_FLAG_MASK)
		turnoffs &= ~COLOR_FLAG_MASK;

//fprintf(stderr, "Turnons = 0x%08x   Turnoffs = 0x%08x\n", turnons, turnoffs);

	while (turnoffs) {
		UINT32	state = get_this_state(get_last_entry(OutputStack));
		
		if (!(state & turnoffs)) {
			// For Colors and Sizes only turn on a color or size
			// if it is asked for
			if (state & ~(SIZE_FLAG_MASK|COLOR_FLAG_MASK))
				turnons |= state;
		}
		turnoffs &= ~state;

		fpfinalcount += strlen(return_end_flag(state));
		fputs(return_end_flag(state), fpfinal);
		
		release_record(OutputStack, get_last_entry(OutputStack));
	}

	// We are turning on a color and the current color is the same don't 
	// turn it on.  The same is for size
	turnons = should_add_color(turnons, OutputStack);
	turnons = should_add_size(turnons, OutputStack);

	if (count_changes(turnons) == 1) {
		fpfinalcount += strlen(return_start_flag(turnons));
		fputs(return_start_flag(turnons), fpfinal);
		new_record_special(OutputStack, turnons);
	} else if (turnons)
		look_ahead(ListEntry->next, turnons, OutputStack);

	fpfinalcount += strlen(ListEntry->doc);
	fputs(ListEntry->doc, fpfinal);
}

void 
print_function(PLIST ListPtr)
{
	PLIST_ENTRY tempPtr  = ListPtr->first;
	PLIST outputStack;

	// Setup
	new_list(&outputStack);

	// Process
	//fprintf(stdout, "\n---START OF OPTIMIZATION---\n");

	while (tempPtr != NULL) {
		process_entry(tempPtr, outputStack);
		tempPtr = get_next_entry(tempPtr);
	}

	tempPtr = get_last_entry(outputStack);
	while (tempPtr != NULL) {
		fpfinalcount += strlen(return_end_flag(get_this_state(tempPtr)));
		fputs(return_end_flag(get_this_state(tempPtr)), fpfinal);
		tempPtr = get_previous_entry(tempPtr);
	}

	//fprintf(stdout, "\n---END OF OPTIMIZATION---\n");
	fflush(stdout);
	fflush(stderr);

	// Cleanup
	release_list(outputStack);
}
