

/* SML/NG Parser */
/* TabStops are set to 4 */

%{

#include "llist.h"

#define YYSTYPE char

extern UINT32 fpintermediatecount;
extern FILE *fpintermediate;

STATIC INLINE
greg_printf(char *string, int len)
{
	fputs(string, fpintermediate);
	fpintermediatecount += len;
}

STATIC INLINE
greg_printc(char c)
{
	fputc(c, fpintermediate);
	fpintermediatecount++;
}

typedef struct pl_state_holder
{
	UINT32	state;
	UINT32	bcount;
	UINT32	icount;
	UINT32	ucount;
	UINT32	scount;
	UINT32	ttcount;
} pl_state_holder;

typedef struct color_state_holder
{
	UINT32	state;
	UINT32	count;
} color_state_holder;

typedef struct size_state_holder
{
	UINT32	state;
	UINT32	count;
} size_state_holder;

PLIST	head_of_list	= NULL;
PLIST	sflag_stack		= NULL;
PLIST	plflag_stack	= NULL;
PLIST	size_stack		= NULL;
PLIST	color_stack		= NULL;

pl_state_holder	*plstate = NULL;
color_state_holder *colorstate = NULL;
size_state_holder *sizestate = NULL;

extern int debug;

STATIC INLINE pl_state_holder *
new_pl_state_holder(void)
{
	pl_state_holder *tmp;

	tmp = (pl_state_holder *)malloc(sizeof(pl_state_holder));
	if (!tmp)
		exit(-1);
	memset(tmp, 0, sizeof(pl_state_holder));
	return tmp;
}

STATIC INLINE void
free_pl_state_holder(pl_state_holder *p)
{
	if (p)
		free(p);
}

STATIC INLINE void
push_pl_state(void)
{
	pl_state_holder *new_state = new_pl_state_holder();
	new_record(plflag_stack, (UINT32)plstate);
	plstate = new_state;
}

STATIC INLINE void
pop_pl_state(void)
{
	PLIST_ENTRY die = get_last_entry(plflag_stack);
	UINT32 pstate = get_this_state(die);

	release_record(plflag_stack, die);
	free_pl_state_holder(plstate);
	plstate = (pl_state_holder *)pstate;
}

STATIC INLINE color_state_holder *
new_color_state_holder(void)
{
	color_state_holder *tmp;

	tmp = (color_state_holder *)malloc(sizeof(color_state_holder));
	if (!tmp)
		exit(-1);
	memset(tmp, 0, sizeof(color_state_holder));
	return tmp;
}

STATIC INLINE void
free_color_state_holder(color_state_holder *p)
{
	if (p)
		free(p);
}

STATIC INLINE void
push_color_state(void)
{
	color_state_holder *new_state = new_color_state_holder();
	new_record(color_stack, (UINT32)colorstate);
	new_state->state = colorstate->state;
	colorstate = new_state;
}

STATIC INLINE void
pop_color_state(void)
{
	PLIST_ENTRY die = get_last_entry(color_stack);
	UINT32 pstate = get_this_state(die);

	release_record(color_stack, die);
	free_color_state_holder(colorstate);
	colorstate = (color_state_holder *)pstate;
}


STATIC INLINE size_state_holder *
new_size_state_holder(void)
{
	size_state_holder *tmp;

	tmp = (size_state_holder *)malloc(sizeof(size_state_holder));
	if (!tmp)
		exit(-1);
	memset(tmp, 0, sizeof(size_state_holder));
	return tmp;
}

STATIC INLINE void
free_size_state_holder(size_state_holder *p)
{
	if (p)
		free(p);
}

STATIC INLINE void
push_size_state(void)
{
	size_state_holder *new_state = new_size_state_holder();
	new_record(size_stack, (UINT32)sizestate);
	new_state->state = sizestate->state;
	sizestate = new_state;
}

STATIC INLINE void
pop_size_state(void)
{
	PLIST_ENTRY die = get_last_entry(size_stack);
	UINT32 pstate = get_this_state(die);

	release_record(size_stack, die);
	free_size_state_holder(sizestate);
	sizestate = (size_state_holder *)pstate;
}

#define	INSERT_SIZE(x, y)	{										\
		PLIST_ENTRY entry = get_last_entry(head_of_list);			\
		UINT32 pstate = get_this_state(entry);						\
																	\
		if ((pstate & SIZE_FLAG_MASK) != (y))						\
		{															\
			greg_printf((x), 3);									\
			if (entry->current != 0)								\
				new_record(head_of_list, get_this_state(entry));	\
			sizestate->state = pstate;								\
			push_size_state();										\
			add_state(head_of_list, (y));							\
		}															\
		sizestate->count++;											\
}


#define	REMOVE_SIZE(x)	{ 										\
	sizestate->count--;											\
	if (sizestate->count == 0)									\
	{															\
		PLIST_ENTRY entry = get_last_entry(head_of_list);		\
																\
		greg_printf((x), 4);									\
																\
		pop_size_state();										\
																\
		if (entry->current != 0)								\
			new_record(head_of_list, sizestate->state);			\
		else entry->state = sizestate->state;					\
	}															\
}

#define	INSERT_COLOR(x, y)	{										\
		PLIST_ENTRY entry = get_last_entry(head_of_list);			\
		UINT32 pstate = get_this_state(entry);						\
																	\
		if ((pstate & COLOR_FLAG_MASK) != (y))						\
		{															\
			greg_printf((x), 3);									\
			if (entry->current != 0)								\
				new_record(head_of_list, get_this_state(entry));	\
			colorstate->state = pstate;								\
			push_color_state();										\
			add_state(head_of_list, (y));							\
		}															\
		colorstate->count++;										\
}

#define	REMOVE_COLOR(x)		{									\
	colorstate->count--;										\
	if (colorstate->count == 0)									\
	{															\
		PLIST_ENTRY entry = get_last_entry(head_of_list);		\
																\
		greg_printf((x), 4);									\
																\
		pop_color_state();										\
																\
		if (entry->current != 0)								\
			new_record(head_of_list, colorstate->state);		\
		else entry->state = colorstate->state;					\
	}															\
}

#define	INSERT_WS(x)	{											\
	if (add_char(head_of_list, (x)) == 1)							\
	{																\
		if (get_this_state(get_last_entry(head_of_list)) & TT_Flag)	\
			greg_printc((x));										\
		else greg_printc(' ');										\
	}																\
}

%}

/* BISON Declarations */
%token TOK_B_START
%token TOK_B_END
%token TOK_EM_START
%token TOK_EM_END
%token TOK_I_START
%token TOK_I_END
%token TOK_PL_START
%token TOK_PL_END
%token TOK_S_START
%token TOK_S_END
%token TOK_TT_START
%token TOK_TT_END
%token TOK_U_START
%token TOK_U_END
%token TOK_0_START
%token TOK_0_END
%token TOK_1_START
%token TOK_1_END
%token TOK_2_START
%token TOK_2_END
%token TOK_3_START
%token TOK_3_END
%token TOK_4_START
%token TOK_4_END
%token TOK_5_START
%token TOK_5_END
%token TOK_6_START
%token TOK_6_END
%token TOK_7_START
%token TOK_7_END
%token TOK_8_START
%token TOK_8_END
%token TOK_9_START
%token TOK_9_END
%token TOK_r_START
%token TOK_r_END
%token TOK_g_START
%token TOK_g_END
%token TOK_b_START
%token TOK_b_END
%token TOK_c_START
%token TOK_c_END
%token TOK_m_START
%token TOK_m_END
%token TOK_y_START
%token TOK_y_END
%token TOK_k_START
%token TOK_k_END
%token TOK_w_START
%token TOK_w_END

%token TOK_CR
%token TOK_LF
%token TOK_TAB
%token TOK_SPC

%token TOK_CHAR

/* Grammar follows */
%%

document:    /* empty string */
	| document TOK_CHAR	{ 
		greg_printc($2);
		add_char(head_of_list, $2);
	}
	| document TOK_TAB	{ INSERT_WS('	'); }
	| document TOK_LF	{ INSERT_WS('\n'); }
	| document TOK_CR	{ INSERT_WS('\r'); }
	| document TOK_SPC	{ INSERT_WS(' '); }

	| document TOK_B_START	{
		PLIST_ENTRY entry = get_last_entry(head_of_list);

		if (plstate->bcount == 0)
		{
			greg_printf("<B>", 3);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			add_state(head_of_list, B_Flag);
		}
		plstate->bcount++;
	}
	| document TOK_EM_START	{
		PLIST_ENTRY entry = get_last_entry(head_of_list);
		UINT32 pstate = get_this_state(entry);

		if (!(pstate & S_Flag))
		{
			greg_printf("<EM>", 4);
			if (entry->current != 0)
				new_record(head_of_list, pstate);
			if (pstate & EM_Flag)
				remove_state(get_last_entry(head_of_list), EM_Flag);
			else add_state(head_of_list, EM_Flag);
		}
	}
	| document TOK_I_START	{ 
		PLIST_ENTRY entry = get_last_entry(head_of_list);

		if (plstate->icount == 0)
		{
			greg_printf("<I>", 3);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			add_state(head_of_list, I_Flag);
		}
		plstate->icount++;
	}
	| document TOK_PL_START	{
		PLIST_ENTRY entry = get_last_entry(head_of_list);
		UINT32 pstate = get_this_state(entry);
		PLIST_ENTRY last;

		greg_printf("<PL>", 4);
		if (entry->current != 0)
			new_record(head_of_list, pstate);

		plstate->state = pstate;
		push_pl_state();

		last = get_last_entry(head_of_list);
		remove_state(last, S_Flag);
		remove_state(last, EM_Flag);
		remove_state(last, I_Flag);
		remove_state(last, B_Flag);
		remove_state(last, TT_Flag);
		remove_state(last, U1_Flag);
		remove_state(last, U2_Flag);
		remove_state(last, U3_Flag);
	}
	| document TOK_S_START	{ 
		PLIST_ENTRY entry = get_last_entry(head_of_list);
		UINT32 pstate = get_this_state(entry);

		if (plstate->scount == 0)
		{
			greg_printf("<S>", 3);

			if (entry->current != 0)
				new_record(head_of_list, pstate);
			new_record(sflag_stack, pstate);
			add_state(head_of_list, S_Flag);
			remove_state(get_last_entry(head_of_list), EM_Flag);
		}
		plstate->scount++;
	}
	| document TOK_TT_START	{ 
		PLIST_ENTRY entry = get_last_entry(head_of_list);

		if (plstate->ttcount == 0)
		{
			greg_printf("<TT>", 4);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			add_state(head_of_list, TT_Flag);
		}
		plstate->ttcount++;
	}
	| document TOK_U_START	{
		plstate->ucount++;
		if (plstate->ucount <= 3)
		{
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("<U>", 3);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			add_state(head_of_list, U1_Flag);
			if (plstate->ucount >= 2)
				add_state(head_of_list, U2_Flag);
			if (plstate->ucount >= 3)
				add_state(head_of_list, U3_Flag);
		}
	}

	| document TOK_0_START	{ INSERT_SIZE("<0>", S0_Flag); }
	| document TOK_1_START	{ INSERT_SIZE("<1>", S1_Flag); }
	| document TOK_2_START	{ INSERT_SIZE("<2>", S2_Flag); }
	| document TOK_3_START	{ INSERT_SIZE("<3>", S3_Flag); }
	| document TOK_4_START	{ INSERT_SIZE("<4>", S4_Flag); }
	| document TOK_5_START	{ INSERT_SIZE("<5>", S5_Flag); }
	| document TOK_6_START	{ INSERT_SIZE("<6>", S6_Flag); }
	| document TOK_7_START	{ INSERT_SIZE("<7>", S7_Flag); }
	| document TOK_8_START	{ INSERT_SIZE("<8>", S8_Flag); }
	| document TOK_9_START	{ INSERT_SIZE("<9>", S9_Flag); }

	| document TOK_r_START	{ INSERT_COLOR("<r>", Cr_Flag); }
	| document TOK_g_START	{ INSERT_COLOR("<g>", Cg_Flag); }
	| document TOK_b_START	{ INSERT_COLOR("<b>", Cb_Flag); }
	| document TOK_c_START	{ INSERT_COLOR("<c>", Cc_Flag); }
	| document TOK_m_START	{ INSERT_COLOR("<m>", Cm_Flag); }
	| document TOK_y_START	{ INSERT_COLOR("<y>", Cy_Flag); }
	| document TOK_k_START	{ INSERT_COLOR("<k>", Ck_Flag); }
	| document TOK_w_START	{ INSERT_COLOR("<w>", Cw_Flag); }

	| document TOK_B_END	{ 
		plstate->bcount--;
		if (plstate->bcount == 0)
		{
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("</B>", 4);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			remove_state(get_last_entry(head_of_list), B_Flag);
		}
	}
	| document TOK_EM_END	{ 
		PLIST_ENTRY entry = get_last_entry(head_of_list);
		UINT32 pstate = get_this_state(entry);

		if (!(pstate & S_Flag))
		{
			greg_printf("</EM>", 5);
			if (entry->current != 0)
				new_record(head_of_list, pstate);
			if (pstate & EM_Flag)
				remove_state(get_last_entry(head_of_list), EM_Flag);
			else add_state(head_of_list, EM_Flag);
		}
	}
	| document TOK_I_END	{ 
		plstate->icount--;
		if (plstate->icount == 0)
		{
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("</I>", 4);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			remove_state(get_last_entry(head_of_list), I_Flag);
		}
	}
	| document TOK_PL_END	{
		PLIST_ENTRY entry = get_last_entry(head_of_list);

		greg_printf("</PL>", 5);

		pop_pl_state();

		if (entry->current != 0)
			new_record(head_of_list, plstate->state);
		else entry->state = plstate->state;
	}
	| document TOK_S_END	{ 
		plstate->scount--;
		if (plstate->scount == 0)
		{
			PLIST_ENTRY die = get_last_entry(sflag_stack);
			UINT32 pstate = get_this_state(die);
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("</S>", 4);

			release_record(sflag_stack, die);

			if (entry->current != 0)
				new_record(head_of_list, pstate);
			else entry->state = pstate;
		}
	}
	| document TOK_TT_END	{ 
		plstate->ttcount--;
		if (plstate->ttcount == 0)
		{
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("</TT>", 5);

			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));
			remove_state(get_last_entry(head_of_list), TT_Flag);
		}
	}
	| document TOK_U_END	{
		if (plstate->ucount <= 3)
		{
			PLIST_ENTRY entry = get_last_entry(head_of_list);

			greg_printf("</U>", 4);
			
			if (entry->current != 0)
				new_record(head_of_list, get_this_state(entry));

			remove_state(get_last_entry(head_of_list), U3_Flag);
			if (plstate->ucount <= 2)
				remove_state(get_last_entry(head_of_list), U2_Flag);
			if (plstate->ucount <= 1)
				remove_state(get_last_entry(head_of_list), U1_Flag);
		}
		plstate->ucount--;
	}

	| document TOK_0_END	{ REMOVE_SIZE("</0>"); }
	| document TOK_1_END	{ REMOVE_SIZE("</1>"); }
	| document TOK_2_END	{ REMOVE_SIZE("</2>"); }
	| document TOK_3_END	{ REMOVE_SIZE("</3>"); }
	| document TOK_4_END	{ REMOVE_SIZE("</4>"); }
	| document TOK_5_END	{ REMOVE_SIZE("</5>"); }
	| document TOK_6_END	{ REMOVE_SIZE("</6>"); }
	| document TOK_7_END	{ REMOVE_SIZE("</7>"); }
	| document TOK_8_END	{ REMOVE_SIZE("</8>"); }
	| document TOK_9_END	{ REMOVE_SIZE("</9>"); }

	| document TOK_r_END	{ REMOVE_COLOR("</r>"); }
	| document TOK_g_END	{ REMOVE_COLOR("</g>"); }
	| document TOK_b_END	{ REMOVE_COLOR("</b>"); }
	| document TOK_c_END	{ REMOVE_COLOR("</c>"); }
	| document TOK_m_END	{ REMOVE_COLOR("</m>"); }
	| document TOK_y_END	{ REMOVE_COLOR("</y>"); }
	| document TOK_k_END	{ REMOVE_COLOR("</k>"); }
	| document TOK_w_END	{ REMOVE_COLOR("</w>"); }
;

%%

int
yyerror(char *s)
{
	fprintf(stderr, "%s\n", s);
}


void
yyparse_init()
{
	new_list(&sflag_stack);
	new_list(&plflag_stack);
	new_list(&size_stack);
	new_list(&color_stack);
	new_list(&head_of_list);
	new_record(head_of_list, 0);
	plstate = new_pl_state_holder();
	colorstate = new_color_state_holder();
	sizestate = new_size_state_holder();
}

void
yyparse_complete()
{
	release_list(sflag_stack);
	release_list(plflag_stack);
	release_list(size_stack);
	release_list(color_stack);

	free_size_state_holder(sizestate);
	free_color_state_holder(colorstate);
	free_pl_state_holder(plstate);
}

